#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"
#include "..\prng.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::makeRabbit(Player *p)
{
	sendPublic("*timereset");
	sendPublic("*timer 10");
	sendPrivate(p, "*prize 5000");

	rabbit = p;
	countdown[0] = 10 * 60;
	countdown[1] = 0;
}

void botInfo::makeRemark(Player *b_new, Player *b_old)
{
	char *old;
	char *name;

	if (!b_new) return;
	else		name = b_new->name;
	if (!b_old) old = "";
	else		old = b_old->name;

	BYTE sound = 0;
	String s;
	s = "*arena ";

	switch (rnd.getNextI() % 13)
	{
	case 0:
		{
			s += "The BountyRabbit(tm) has struck again! Zone in panic! ";
			s += name;
			s += " now at large! $5000 reward!";
			sound = SND_Churchill;
		}
		break;
	case 1:
		{
			s += "Hop. hop. hop. Another rabbit bit the dust. ";
			s += name;
			s += " now holds the 5000 bounty!";
			sound = SND_Inconceivable;
		}
		break;
	case 2:
		{
			s += old;
			s += "(old rabbit) kb ";
			s += name;
			s += "(new rabbit with 5000 bounty)";
			sound = SND_Ahhhh;
		}
		break;
	case 3:
		{
			s += "Uh oh, ";
			s += name;
			s += " is growing ears! He's contracted the BountyRabbit(tm) virus";
			if (*old)
			{
				s += " from ";
				s += old;
			}
			s += "!";
			sound = SND_Scream;
		}
		break;
	case 4:
		{
			s += name;
			s += " is now the BountyRabbit! Type ?target=5000 for a Big Light Blue indicator on your radar";
			sound = SND_Ooooo;
		}
		break;
	case 5:
		{
			s += name;
			s += " has been rabbitified! Type ?time to check how much time is left until he wins";
			sound = SND_Geeee;
		}
		break;
	case 6:
		{
			s += "Another waskalie wabbit has been sighted: ";
			s += name;
			s += ". And it's wabbit-hunting season!";
			sound = SND_Gibberish;
		}
		break;
	case 7:
		{
			s += name;
			s += " has vanquished the villainous old rabbit! He has been granted a $5000 reward";
			sound = SND_Reagan;
		}
		break;
	case 8:
		{
			s += "A new bounty rabbit has been born and it ain't pretty! ";
			s += name;
			s += " holds the sucker";
			sound = SND_Crying;
		}
		break;
	case 9:
		{
			s += "Uh oh, ";
			s += name;
			s += " is growing ears! He's contracted the BountyRabbit(tm) virus from ";
			s += old;
			s += "!";
			sound = SND_Discretion;
		}
		break;
	case 10:
		{
			s += "The BountyRabbit(tm) has struck again! Zone in panic! ";
			s += name;
			s += " now at large! $5000 reward!";
			sound = SND_WorldUnderAttack;
		}
		break;
	case 11:
		{
			s += "Zee rebbeet, ";
			s += name;
			s += ", veell be-a in zee stoo suun inuoogh. $5000 preece-a teg";
			sound = SND_Ohhhh;
		}
		break;
	case 12:
		{
			String killer = name;
			s += "Psst! ";
			s += killer.mid(1, 32);
			s += killer.left(1);
			s += "ay has ethay OuntyRabbitbay irusvay! Asspay it onyay!";
			sound = SND_Phone;
		}
		break;
	default:
		{
			s += "New bountyrabbit: ";
			s += name;
			sound = SND_PlayOnce;
		}
		break;
	};

	sendPublic(sound, s);
}

Player *botInfo::chooseRabbit()
{
	int max = getIngame();
	if (max < 2) return 0;

	int sel = rnd.getNextI() % max;

	_listnode <Player> *parse = playerlist->head;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
		{
			if (!(sel--))
				return p;
		}

		parse = parse->next;
	}

	return 0;
}

int botInfo::getIngame()
{
	_listnode <Player> *parse = playerlist->head;

	int total = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
			++total;

		parse = parse->next;
	}

	return total;
}

Sint32 getNumeric(Sint32 y)
{
	return (y * 20 / TILE_MAX_Y) + 1;
}

char getAlpha(Sint32 x)
{
	return 'A' + char(x * 20 / TILE_MAX_X);
}

String getCoords(Sint32 x, Sint32 y)
{
	x >>= 4;
	y >>= 4;

	String s;

	char alpha[2];
	alpha[0] = getAlpha(x);
	alpha[1] = '\0';

	s += alpha;
	s += getNumeric(y);

	return s;
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (rabbit)
			switch (countdown[0])
			{
			case 0:
				{
					String s;
					s += "*arena ";
					s += rabbit->name;
					s += " has survived the 10-minute duration as Bounty Rabbit!  Congratulations on becoming an Uberabbit!";
					sendPublic(SND_Hallellula, s);

					rabbit = 0;
					Player *xp = chooseRabbit();

					if (xp)
					{
						String s;
						s += "*arena New rabbit: ";
						s += xp->name;
						sendPublic(s);

						makeRabbit(xp);
					}
				}
				break;
			case 60:
				{
					String s;
					s += "*arena ";
					s += rabbit->name;
					s += " has survived for 9 minutes as Bounty Rabbit!  Last seen at ";
					s += getCoords(rabbit->pos.x, rabbit->pos.y);
					sendPublic(SND_Scream, s);
				}
				break;
			case 300:
				{
					String s;
					s += "*arena ";
					s += rabbit->name;
					s += " has survived for 5 minutes as Bounty Rabbit!  Last seen at ";
					s += getCoords(rabbit->pos.x, rabbit->pos.y);
					sendPublic(SND_Scream, s);
				}
				break;
			};

			if (countdown[1] == 0)
			{
				if (rabbit)
				{
					String s;
					s += "*arena ";
					s += rabbit->name;
					s += " overstayed his welcome in the safety zone.";
					sendPublic(s);

					rabbit = 0;
					Player *xp = chooseRabbit();

					if (xp)
					{
						String s;
						s += "*arena New rabbit: ";
						s += xp->name;
						sendPublic(s);

						makeRabbit(xp);
					}
				}
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			rabbit = 0;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];

			if (rabbit)
				sendPrivate(rabbit, "*prize 5000");
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];

			if (p == rabbit)
			{
				if (p->safety != 0)
				{
					if (countdown[1] < 0) countdown[1] = 20;
				}
				else
				{
					countdown[1] = 0;
				}
			}
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];

			if (p == rabbit)
			{
				rabbit = 0;
				makeRemark(k, p);
				makeRabbit(k);
			}
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];

			if (p == rabbit)
				sendPrivate(p, "*prize 5000");
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];

			if (p == rabbit)
			{
				rabbit = 0;
				Player *xp = chooseRabbit();

				if (xp)
				{
					String s;
					s += "*arena New rabbit: ";
					s += xp->name;
					s += p->ship;
					sendPublic(s);

					makeRabbit(xp);
				}
			}
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];

			if (p == rabbit)
				sendPrivate(p, "*prize 5000");
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			if (p == rabbit)
			{
				rabbit = 0;
				Player *xp = chooseRabbit();

				if (xp)
				{
					String s;
					s += "*arena New rabbit: ";
					s += xp->name;
					sendPublic(s);

					makeRabbit(xp);
				}
			}
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major != CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  Major DLL version mismatch!"));

				CONNECTION_DENIED = true;

				return;
			}
			else if (minor != CORE_MINOR_VERSION)
				tell(makeEcho("DLL plugin connected.  Minor DLL version mismatch!"));
			else
				tell(makeEcho("DLL plugin connected."));
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	default:
		{
			String s;
			s += "Unknown message from bot (";
			s += event.code;
			s += ")  Version mismatch?";
			tell(makeEcho(s));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
